from __future__ import annotations

import re
from typing import TYPE_CHECKING

from ..boxes import ModDBList, Option, Thumbnail
from ..enums import ThumbnailType
from ..utils import concat_docs, get_date, get_list_stats, join
from .base import BaseMetaClass

if TYPE_CHECKING:
    from bs4 import BeautifulSoup


class ReviewList(ModDBList):
    """Represents a list of reviews

    Attributes
    -----------
    current_page : int
        The page of results this objects represents
    total_pages : int
        The total amount of result pages available
    total_results : int
        The total amount of results available
    """

    def _parse_method(self, html: BeautifulSoup):
        return parse_reviews(html)


def parse_reviews(html):
    review_box = html.find("div", class_="normalbox browsebox")

    try:
        raw_ratings = (
            review_box.find("div", class_="inner")
            .find("div", class_="table")
            .find_all("div", class_=["rowrating"], recursive=False)
        )
    except AttributeError:
        return [], 1, 1, 0

    reviews = []
    for rating in raw_ratings:
        next_sibling = rating.find_next_sibling("div", class_=["rowcontent"])
        if next_sibling is not None and "rowcontentnext" in next_sibling["class"]:
            reviews.append(Review(rating=rating, text=next_sibling))
        else:
            reviews.append(Review(rating=rating))

    return reviews, *get_list_stats(review_box, 10)


@concat_docs
class Review:
    """Represents a review.

    Filtering
    -----------
    rating : int
        A value from 1 to 10 denoting the rating number you're looking for
    sitearea : Category
        The type of model the rating is for (mod, engine, game)

    Sorting
    --------
        * **ratingalt** - rating number, desc is biggest to lowest, asc is lowest to biggest
        * **memberipid** - sort reviewer account age, asc is oldest reviewer first
        * **positive** - how many people agree with it, desc is most to least people agreeing
        * **negative** - how many people disagree with it, desc is most to least people disagreeing
        * **id** - when it was added to moddb, asc is oldest, desc is most recent

    Attributes
    -----------
    id : int
        The review id
    text : str
        The contents of the review. Can be none if the member hasn't left any
    rating : int
        An int out of 10 representing the rating left with this review.
    author : Thumbnail
        A member like thumbnail of the member who left the review
    date : datetime.datetime
        Date and time of the review creation
    agree : str
            Link to agree with the review
    disagree : str
        Link to disagree with the review
    """

    def __init__(self, **attrs):
        text = attrs.get("text")
        if text:
            self.text = text.text
        else:
            self.text = None

        review = attrs.get("rating")
        self.rating = int(review.span.string)

        # id and hash are none if the review doesn't have content
        try:
            strings = ("Agree", "Delete", "Disagree")
            self.id = int(
                re.findall(r"siteareaid=(\d*)", review.find("a", title=strings)["href"])[0]
            )
        except TypeError:
            self.id = None

        try:
            self._hash = re.findall(r"hash=(.*)&", review.find("a", title="Delete")["href"])[0]
        except TypeError:
            self._hash = None

        author = review.div.a
        self.author = Thumbnail(
            url=author["href"],
            name=author.string.split(" ")[0],
            type=ThumbnailType.member,
        )
        self.date = get_date(review.div.span.time["datetime"])

        try:
            self.agree = join(review.find("a", title="Agree")["href"])
            self.disagree = join(review.find("a", title="Disagree")["href"])
        except TypeError:
            self.agree = None
            self.disagree = None

    def __repr__(self):
        return f"<Review author={self.author.name} rating={self.rating}>"


@concat_docs
class Poll(BaseMetaClass):
    """Represents a poll. Cannot be voted for due to restrictions implemented by the website.

    Parameters
    -----------
    html : bs4.BeautifulSoup
        The html to parse. Allows for finer control.

    Filtering
    ----------
        month : Month
            The month the poll you're looking for should be from
        year : int
            A int representing a year between 2002 and now. Anything below or above 2002 will
            always return zero results.

    Sorting
    --------
        * **totalvotes** - how many people voted on the poll, desc is most to least
        * **name** - sort the poll alphabetically by name, asc is a-z
        * **date** - when it was added to moddb, asc is oldest, desc is most recent

    Attributes
    -----------
    question : str
        The question of the poll
    author : Thumbnail
        A member like thumbnail of the member who posted the poll, usually ModDB staff
    total : int
        The total number of votes that have been cast
    options : List[Option]
        The list of available options for the poll
    """

    def __init__(self, html: BeautifulSoup):
        poll = html.find("div", class_="poll")
        self.question = (
            poll.parent.parent.parent.find("div", class_="normalcorner")
            .find("span", class_="heading")
            .string
        )
        self.name = self.question
        super().__init__(html)
        author = poll.find("p", class_="question").find("a")
        self.author = Thumbnail(name=author.string, url=author["href"], type=ThumbnailType.member)

        self.total = int(
            re.search(r"([\d,]*) votes", poll.find("p", class_="question").text)[1].replace(",", "")
        )

        percentage = poll.find_all("div", class_="barouter")
        rest = poll.find_all("p")[1:]

        self.options = []
        for index, _ in enumerate(percentage):
            raw = percentage[index].div.string.replace("%", "").replace("\xa0", "")
            percent = float(f"0.{raw}")
            text = re.sub(r"\([\d,]* vote(s)?\)", "", rest[index].text)
            votes = int(
                re.search(r"([\d,]*) vote(s)?", rest[index].span.string)[1].replace(",", "")
            )
            self.options.append(Option(text=text, votes=votes, percent=percent))

    def __repr__(self):
        return f"<Poll question={self.question}>"



import datetime
import re
import sys
from typing import BinaryIO, List

import bs4
import requests

from ..boxes import Mirror, Thumbnail
from ..enums import AddonCategory, FileCategory, MediaCategory, ThumbnailType
from ..utils import (
    BASE_URL,
    concat_docs,
    get_date,
    get_page,
    get_views,
    join,
    prepare_request,
)
from .base import BaseMetaClass


def parse_location(html) -> list[Thumbnail] | None:
    location = html.find("h5", string="Location").parent.find_all("a")
    if location is None:
        return None

    return [
        Thumbnail(
            type=ThumbnailType[location[x].string.lower()[:-1]],
            url=location[x + 1]["href"],
            name=location[x + 1].string,
        )
        for x in range(0, len(location) - 1, 2)
    ]


@concat_docs
class File(BaseMetaClass):
    """An oject representing a file on ModDB, a file is something posted by the page owner which is directly linked
    to the page. It is endorsed by the page owner and they should do everythign they can to make sure that it is safe.
    As compared to an addon that may be added by fans to the page and that are files meant to work with the page but
    that are not directly related to the page. E.x the file of a mod page would be the mod files used to install the
    mod whereas an addon could be something like a fan-made texture pack for the mod or a map.

    Parameters
    -----------
    html : bs4.BeautifulSoup
        The html to parse. Allows for finer control.

    Filtering
    ----------
    category  : :class:`.FileCategory`
        The type of file (audio, video, demo, full version....)
    categoryaddon : :class:`.AddonCategory`
        The type of addon (map, textures, ect...)
    game : Union[:class:`.Game`, :class:`.Object`]
        An game object or an object with an id attribute which represents the
        game the file belongs to.
    timeframe : :class:`.TimeFrame`
        The time period this was released in (last 24hr, last week, last month)

    Sorting
    --------
        * **released** - when the object was released, asc is oldest, desc is most recent
        * **id** - when it was added to moddb, asc is oldest, desc is most recent
        * **ranktoday** - order by daily ranking, asc is highest ranked, desc is lowest rank
        * **visitstotal** - order by most views, asc is highest views, desc is lowest views
        * **rating** - order by rating, asc is highest rating, desc is lowest rating
        * **name** - order alphabetically, asc is a-z, desc is z-a
        * **date** - order by upload date, asc is most recent first, desc is oldest first

    Attributes
    -----------
    filename : str
        The name of the file
    hash : str
        The MD5 hash of the file
    name : str
        The name of the page
    size : int
        the file size in bytes
    today : int
        The number of downloads today
    downloads : int
        The total number of times this file has been downloaded
    category : FileCategory
        The category of the file
    author : Thumbnail
        A member type thumbnail of the member who uploaded the file
    date : datetime.datetime
        The date the file was uploaded
    button : str
        html code for the embed button
    widget : str
        html code for the embed widget
    description : str
        Description of the file, as written by the author
    preview : str
        URL of the preview image for the file
    location: list[Thumbnail]
        An ordered list detailing the hierarchy of entities the
        file or addon sits under. The last one being the entity
        directly attached to this file.
    """

    def __init__(self, html: bs4.BeautifulSoup):
        if html.find("span", string="File Deleted", class_="heading"):
            raise ValueError("This file has been removed")

        info = html.find("div", class_="table tablemenu")
        file = {
            x.string.lower(): x.parent.span.string.strip()
            for x in info.find_all("h5", string=("Filename", "Size", "MD5 Hash"))
        }
        self.name = (
            html.find("a", title="Report").parent.parent.find("span", class_="heading").string
        )
        self.filename = file["filename"]
        super().__init__(html)

        self.hash = file["md5 hash"]
        self.size = int(re.sub(r"[(),bytes]", "", file["size"].split(" ")[1]))

        downloads = html.find("h5", string="Downloads").parent.a.string
        self.today = int(re.sub(r"[(),today]", "", downloads.split(" ")[1]))
        self.downloads = int(downloads.split(" ")[0].replace(",", ""))

        try:
            self.category = FileCategory(
                int(info.find("h5", string="Category").parent.a["href"].split("=")[-1])
            )
        except ValueError:
            self.category = AddonCategory(
                int(info.find("h5", string="Category").parent.a["href"].split("=")[-1])
            )

        uploader = info.find("h5", string="Uploader").parent.a
        self.author = Thumbnail(
            url=uploader["href"], name=uploader.string, type=ThumbnailType.member
        )

        self.date = get_date(info.find("h5", string="Added").parent.span.time["datetime"])
        self.button = info.find("h5", string="Embed Button").parent.span.input["value"]
        self.widget = info.find("h5", string="Embed Widget").parent.span.input["value"]

        self.description = html.find("p", id="downloadsummary").string

        self.preview = html.find_all("img", src=True)[0]["src"]

        self.location = parse_location(html)

    def __repr__(self):
        return f"<{self.__class__.__name__} name={self.name} type={self.category.name}>"

    def save(self, file_obj: BinaryIO, *, mirror=None, chunk_size: int = 10_000_000):
        """Save the file to an object. This functions makes
        two requests. If you pass a valid mirror it will
        make only one request.

        Parameters
        -----------
        file_obj : BinaryIO
            The file obj to save the file to. The binary data
            will be streamed to that object.
        mirror : Optional[Mirror]
            An optional mirror object to download the
            file from a specific moddb mirror
        chunk_size: int
            The size of the chunks to stream the response
            back in. 10MB by default
        """
        if mirror is None:
            download = get_page(f"{BASE_URL}/downloads/start/{self.id}")
            url = download.find("a", string=f"download {self.filename}")["href"]
        else:
            url = mirror._url

        SESSION: requests.Session = sys.modules["moddb"].SESSION
        prepped = prepare_request(requests.Request("GET", join(url)), SESSION)
        with SESSION.send(prepped, stream=True) as r:
            r.raise_for_status()
            for chunk in r.iter_content(chunk_size=chunk_size):
                file_obj.write(chunk)

    def get_mirrors(self) -> List[Mirror]:
        """Get all the mirrors from which a file can be downloaded. This
        can then be passed to File.save to download from a specific mirror.


        Returns
        --------
        List[Mirror]
            A list of Mirror objects"""

        html = get_page(f"https://www.moddb.com/downloads/start/{self.id}/all")
        mirrors_div = html.find("div", class_="mirrors").find_all("div", recursive=False)
        mirrors = []
        for mirror in mirrors_div:
            mirror_match = re.match(
                r"(.*) #([0-9]*) \((\w+), (\w+)\)", mirror.div.p.contents[-1].strip()
            )
            stats_match = re.match(
                r"([0-9,]*) downloads? served, ([0-9.]*)% capacity",
                mirror.div.span.string,
            )

            mirrors.append(
                Mirror(
                    name=mirror_match.group(1),
                    index=int(mirror_match.group(2)),
                    city=mirror_match.group(3),
                    country=mirror_match.group(4),
                    served=int(stats_match.group(1).replace(",", "")),
                    capacity=float(stats_match.group(2)),
                    url=mirror.div.p.a["href"],
                )
            )

        return mirrors


@concat_docs
class Addon(File):
    """Object representing an addon. Seemingly the only difference between an addon and a file is in
    the semantics. A file often represents something official released by the page, e.g. the mod installation
    or an official guide where as addons are often fan made and might not be directly endorsed by the page owners
    even if it is allowed. They literally add on to the page's content without becoming part of it. There is a slight
    difference in their profiles but nothing beyond that.

    Parameters
    -----------
    html : bs4.BeautifulSoup
        The html to parse. Allows for finer control.

    Filtering
    ----------
    categoryaddon : :class:`.AddonCategory`
        The type of addon (map, textures, ect...)
    licence : :class:`.Licence`
        The licence of the addon
    game : Union[:class:`.Game`, :class:`.Object`]
        An game object or an object with an id attribute which represents the
        game the addon belongs to.
    timeframe : :class:`.TimeFrame`
        The time period this was released in (last 24hr, last week, last month)

    Sorting
    --------
        * **released** - when the object was released, asc is oldest, desc is most recent
        * **id** - when it was added to moddb, asc is oldest, desc is most recent
        * **ranktoday** - order by daily ranking, asc is highest ranked, desc is lowest rank
        * **visitstotal** - order by most views, asc is highest views, desc is lowest views
        * **rating** - order by rating, asc is highest rating, desc is lowest rating
        * **name** - order alphabetically, asc is a-z, desc is z-a
        * **licence** - order based on licence
        * **date** - order by upload date, asc is most recent first, desc is oldest first


    """

    pass


@concat_docs
class Media(BaseMetaClass):
    """Represents an image, audio file or video file on

    Parameters
    -----------
    html : bs4.BeautifulSoup
        The html to parse. Allows for finer control.

    Filtering
    -----------
    sitearea : :class:`.Category`
        The type of model the media belongs to. Category.downloads is not valid for this.

    Sorting
    --------
        * **ranktoday** - order by daily ranking, asc is highest ranked, desc is lowest rank
        * **visitstotal** - order by most views, asc is highest views, desc is lowest views
        * **name** - order alphabetically, asc is a-z, desc is z-a
        * **id** - order by upload date, asc is most recent first, desc is oldest first

        Exclusive to videos and audios
        * **duration** - order by duration, asc is shortest to longest, desc is longest first

    Attributes
    -----------
    date : datetime.datetime
        The date the media was uploaded
    name : str
        The name of the media
    author : Thumbnail
        Member type thumbnail of the media uploader
    duration : datetime.timedelta
        Duration of the media in seconds, 0 if it's an image
    size : int
        Size of the files in bytes
    views : int
        Total amount of views
    today : int
        Amount of views today
    filename : str
        The name of the file for the media
    fileurl : str
        The url of the file for the media
    category : MediaCategory
        Whether the media is an image, a video or an audio
    description : str
        The description of the file as given by the file uploader.
    """

    def __init__(self, html: bs4.BeautifulSoup):
        try:
            self.name = html.find("meta", itemprop="name")["content"]
        except TypeError:
            self.name = html.find("img", id="mediaimage")["title"]

        super().__init__(html)
        medias = html.find_all("h5", string=("Date", "By", "Duration", "Size", "Views", "Filename"))
        raw_media = {media.string.lower(): media.parent for media in medias}

        self.date = get_date(raw_media["date"].span.time["datetime"])

        author = raw_media["by"].span.a
        self.author = Thumbnail(
            url=author["href"], name=author.string.strip(), type=ThumbnailType.member
        )

        if "duration" in raw_media:
            duration = raw_media["duration"].span.time.string.strip().split(":")
            duration.reverse()
            times = ["seconds", "minutes", "hours"]
            self.duration = datetime.timedelta(
                **{times[duration.index(x)]: int(x) for x in duration}
            )
        else:
            self.duration = 0

        if "size" in raw_media:
            self.size = tuple(raw_media["size"].span.string.strip().split("×"))

        self.views, self.today = get_views(raw_media["views"].a.string)
        media_player = html.find("video", id="mediaplayer")

        if not media_player:
            self.category = MediaCategory.image
            self.fileurl = html.find("meta", property="og:image")["content"]
        else:
            self.fileurl = media_player.source["src"]
            if "audio" in media_player.source["type"]:
                self.category = MediaCategory.audio
            else:
                self.category = MediaCategory.video

        if "filename" in raw_media:
            self.filename = raw_media["filename"].span.string.strip()
        else:
            self.filename = self.fileurl.split("/")[-1]

        self.description = html.find("meta", {"name": "description"})["content"]

    def __repr__(self):
        return f"<Media name={self.name} type={self.category.name}>"

    def save(self, file_obj: BinaryIO, *, chunk_size: int = 10_000_000):
        """Save the media to an object.

        Parameters
        -----------
        file_obj : BinaryIO
            The file obj to save the file to. The binary data
            will be streamed to that object.
        chunk_size: int
            The size of the chunks to stream the response
            back in. 10MB by default

        """
        SESSION: requests.Session = sys.modules["moddb"].SESSION
        prepped = prepare_request(requests.Request("GET", self.fileurl), SESSION)

        with SESSION.send(prepped, stream=True) as r:
            r.raise_for_status()
            for chunk in r.iter_content(chunk_size=chunk_size):
                file_obj.write(chunk)



from .article import Article, Blog
from .engine import Engine
from .entity import Group, Member, Team
from .file import Addon, File, Media
from .fp import FrontPage
from .game import Game
from .job import Job
from .mod import Mod
from .opinion import Poll, Review, ReviewList
from .platform import Platform
from .ware import Hardware, Software

__all__ = [
    "Mod",
    "Game",
    "Engine",
    "File",
    "Addon",
    "Media",
    "Article",
    "Team",
    "Group",
    "Job",
    "Blog",
    "Member",
    "FrontPage",
    "Review",
    "Platform",
    "Poll",
    "Software",
    "Hardware",
    "ReviewList",
]



from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from ..boxes import PlatformStatistics, Thumbnail
from ..enums import ThumbnailType
from ..utils import LOGGER, concat_docs, get_date, join
from .base import BaseMetaClass
from .mixins import GetEnginesMixin, GetGamesMixin, GetModsMixin, GetWaresMixin

if TYPE_CHECKING:
    from bs4 import BeautifulSoup


@concat_docs
class Platform(
    BaseMetaClass,
    GetModsMixin,
    GetGamesMixin,
    GetEnginesMixin,
    GetWaresMixin,
):
    """Represents the platform supporting the game/engines. Game and engines may have mutiple platforms.

    Parameters
    -----------
    html : bs4.BeautifulSoup
        The html to parse. Allows for finer control.

    Filtering
    ----------
    released : :class:`.Status`
        Current status of release (unreleased, early access, ect...)

    Sorting
    --------
        * **released** - when the object was released, asc is oldest, desc is most recent
        * **id** - when it was added to moddb, asc is oldest, desc is most recent
        * **porder** - what plaform it's on???
        * **company** - company that made it in alphabetical order


    Attributes
    -----------
    name : str
        The name of the platform
    name_id : str
        The name_id of the member, cannot be changed, it's a mix of the original username lowercased with
        spaces removed and shortened.
    id : int
        The moddb id of the platform
    url : str
        The url of the platform page
    description : str
        Description of the platform
    company : Thumbnail
        A team like thumbnail of the company.
    homepage : str
        Link to the homepage of the engine
    date : datetime.datetime
        Date the engine was published
    stats : PlatformStatistics
        Stat data on the platform
    share : dict{str : str}
        Share link of the platform with the name of the share as key and link of the share as url.
    comments : CommentList[Comment]
        Comments on this page
    games : List[Thumbnail]
        A list of games suggested on the platform main page.
    hardware : List[Thumbnail]
        A list of hardware suggested on the platform main page.
    software : List[Thumbnail]
        A list of software suggested on the platform main page.
    engines : List[Thumbnail]
        A list of engines suggested on the platform main page.
    mods : List[Thumbnail]
        A list of mods suggested on the platform main page.
    """

    def __init__(self, html: BeautifulSoup):
        self.name = html.find("a", itemprop="mainEntityOfPage").string
        self.id = None

        self.url = join(html.find("a", itemprop="mainEntityOfPage")["href"])
        self.name_id = self.url.split("/")[0]
        try:
            self.description = html.find("div", id="profiledescription").p.string
        except AttributeError:
            self.description = html.find("p", itemprop="description").string

        try:
            company = html.find("h5", string="Company").parent.span.a
            self.company = Thumbnail(
                name=company.string, url=company["href"], type=ThumbnailType.team
            )
        except AttributeError:
            LOGGER.info(
                "Platform '%s' has no company", self.name, exc_info=LOGGER.level >= logging.DEBUG
            )
            self.company = None

        self.homepage = html.find("h5", string="Homepage").parent.span.a["href"]

        self.date = get_date(html.find("time", itemprop="releaseDate")["datetime"])

        self.stats = PlatformStatistics(html)

        try:
            share = html.find("h5", string="Share").parent.span.find_all("a")
            self.share = {
                "reddit": share[0]["href"],
                "mail": share[1]["href"],
                "twitter": share[2]["href"],
                "facebook": share[3]["href"],
            }
        except (AttributeError, IndexError):
            LOGGER.info(
                "Something funky about share box of platform '%s'",
                self.name,
                exc_info=LOGGER.level >= logging.DEBUG,
            )
            self.share = None

        self.comments = self._get_comments(html)

        hardware = (
            html.find("span", string=" Hardware")
            .parent.parent.parent.find("div", class_="table")
            .find_all("div", recursive=False)[:-1]
        )
        self.hardware = [
            Thumbnail(
                name=x.a["title"],
                url=x.a["href"],
                type=ThumbnailType.hardware,
                image=x.a.img["src"],
            )
            for x in hardware
        ]

        software = (
            html.find("span", string=" Software")
            .parent.parent.parent.find("div", class_="table")
            .find_all("div", recursive=False)[:-1]
        )
        self.software = [
            Thumbnail(
                name=x.a["title"],
                url=x.a["href"],
                type=ThumbnailType.software,
                image=x.a.img["src"],
            )
            for x in software
        ]

        engines = (
            html.find("span", string=" Engines")
            .parent.parent.parent.find("div", class_="table")
            .find_all("div", recursive=False)[:-1]
        )
        self.engines = [
            Thumbnail(
                name=x.a["title"],
                url=x.a["href"],
                type=ThumbnailType.engine,
                image=x.a.img["src"],
            )
            for x in engines
        ]

        mods = (
            html.find("span", string=" Mods")
            .parent.parent.parent.find("div", class_="table")
            .find_all("div", recursive=False)[:-1]
        )
        self.mods = [
            Thumbnail(
                name=x.a["title"],
                url=x.a["href"],
                type=ThumbnailType.mod,
                image=x.a.img["src"],
            )
            for x in mods
        ]

        games = (
            html.find("span", string=" Games")
            .parent.parent.parent.find("div", class_="table")
            .find_all("div", recursive=False)[:-1]
        )
        self.games = [
            Thumbnail(
                name=x.a["title"],
                url=x.a["href"],
                type=ThumbnailType.game,
                image=x.a.img["src"],
            )
            for x in games
        ]

    def __repr__(self):
        return f"<Platform name={self.name}>"



class ModdbException(Exception):
    """Standard exception raised when the moddb.Client encounters an error related to the request
    being made"""

    pass


class AwaitingAuthorisation(ModdbException):
    """This object is awaiting authorisation and can therefore not be accessed at the
    present time."""

    pass


class Ratelimited(ModdbException):
    """You are being ratelimited internally by the library, please be respectful of the
    website and its creators."""

    def __init__(self, message, remaining) -> None:
        super().__init__(message)

        self.remaining = remaining


class AuthError(ModdbException):
    """The user you are trying to login with requires 2FA to login. Use
    the TwoFactorAuthClient object to do so.
    """

    pass



import logging
import re
from typing import List

from bs4 import BeautifulSoup

from ..boxes import (
    CommentList,
    PartialArticle,
    PartialTag,
    Profile,
    ResultList,
    Statistics,
    Style,
    Thumbnail,
    _parse_comments,
    _parse_results,
)
from ..enums import SearchCategory, ThumbnailType
from ..utils import LOGGER, get_page, get_page_type, join
from .mixins import GetTagsMixin, GetWatchersMixin, RSSFeedMixin, SharedMethodsMixin


class BaseMetaClass:
    """An abstract class that implements the attributes present on nearly every page. In addition, it implements
    some shared hidden methods and the top level get_comments method.

    Parameters
    -----------
    html : BeautifulSoup
        The html containing the comments

    Attributes
    -----------
    id : int
        The id of the page
    name_id : str
        The name_id of the entity, cannot be changed, it's a mix of the original username lowercased with
        spaces removed and shortened.
    url : str
        The url of the page
    comments : CommentList[Comment]
        The comments scrapped on this list in order.
    report : str
        URL to report the page
    """

    def __init__(self, html: BeautifulSoup):
        if not getattr(self, "name", None):
            try:
                self.name = html.find("a", itemprop="mainEntityOfPage").string
            except AttributeError:
                self.name = html.find("meta", property="og:title")["content"]

        for index, func in enumerate(
            [
                lambda: int(
                    re.search(r"siteareaid=(\d*)", html.find("a", class_=["reporticon"])["href"])[1]
                ),
                lambda: int(html.find("input", attrs={"name": "siteareaid"})["value"]),
                lambda: int(html.find("meta", property="og:image")["content"].split("/")[-2]),
                lambda: re.findall(
                    r"https:\/\/www\.moddb\.com\/html\/scripts\/autocomplete\.php\?a=mentions&p=home&l=6&u=(\d*)",
                    str(html),
                )[0],
            ]
        ):
            try:
                self.id = func()
                break
            except (AttributeError, TypeError) as e:
                LOGGER.warning(
                    "Failed to get id from method %s for member %s: %s", index, self.name, e
                )
        else:
            raise AttributeError(f"Failed to get id from member {self.name}")

        try:
            self.url = html.find("meta", property="og:url")["content"]
        except TypeError:
            self.url = join(html.find("a", string=self.name)["href"])

        self.name_id = self.url.split("/")[-1]

        try:
            self.report = join(html.find("a", string="Report")["href"])
        except (TypeError, AttributeError):
            self.report = None
            LOGGER.info(
                "'%s' '%s' cannot be reported",
                self.__class__.__name__,
                self.name,
                exc_info=LOGGER.level >= logging.DEBUG,
            )

        self.comments = self._get_comments(html)

        self._html = html

    def __repr__(self):
        return f"<{self.__class__.__name__} name={self.name}>"

    def _get_comments(self, html: BeautifulSoup) -> CommentList:
        """Extracts the comments from an html page and adds them to a CommentList.

        Parameters
        -----------
        html : BeautifulSoup
            The html containing the comments

        Returns
        --------
        CommentList
            The list of parsed comments
        """
        results, current_page, total_pages, total_results = _parse_comments(html)

        return CommentList(
            results=results,
            current_page=current_page,
            total_pages=total_pages,
            url=self.url,
            total_results=total_results,
        )

    def _get(self, url: str, *, params: dict = {}) -> ResultList:
        """This function takes a list of thumbnails of `object_type` in html and returns
        a list of Thumbnail of that object type.

        Parameters
        -----------
        url : str
            The url with the list
        object_type : ThumbnailType
            The type of objects to be expected. Easier to pass and hardcode then to guess
        params : dict
            A dictionnary of filters to pass on to the soup function used to filter the results.

        Returns
        -------
        ResultList[Thumbnail]
            The list of objects present on the page as a list of thumbnails.
        """
        html = get_page(url, params=params)
        results, current_page, total_pages, total_results = _parse_results(html)

        return ResultList(
            results=results,
            params=params,
            url=url,
            current_page=current_page,
            total_pages=total_pages,
            total_results=total_results,
        )

    def _get_media(self, index: int, *, html) -> List[Thumbnail]:
        """Hidden method used to parse media content from the page. Since the only difference is that pages
        with videos have one extra script that tells the window not do any sort of video playing assistance.
        Also used to cached all media on a page.

        Parameters
        ----------
        index : int
            The index of the script containing the cached objects. 1 if it's only images, 2 if videos are
            also included

        Returns
        --------
        List[Thumbnail]
            List of media like thumbnails that can be parsed individually. Can be a very long list.
        """
        script = html.find_all("script", string=True)[index]
        regex = r'new Array\(0, "(\S*)", "(\S*)"'
        matches = re.findall(regex, script.text)

        name_finder = r"\/([a-z0-9-]*)#imagebox"
        return [
            Thumbnail(
                name=re.search(name_finder, match[0])[1],
                url=match[0],
                type=ThumbnailType.media,
                image=match[1],
            )
            for match in matches
        ]

    def get_comments(self, index: int = 1, *, show_deleted=False) -> CommentList:
        """Used to get comments on the model regardless of what page they may be present in. The function
        itself simply relies on two other to make the request and parse the table.

        Parameters
        ----------
        index : int
            The page of the model to get the comments for.
        show_deleted : Optional[bool]
            Pass true to show deleted user comments. Only works if it is a page
            you have permissions on.

        Returns
        --------
        CommentList[Comment]
            A list-like object containing the comments and additional methods
        """
        params = {"deleted": "t" if show_deleted else None}
        return self._get_comments(get_page(f"{self.url}/page/{index}", params=params))


class PageMetaClass(
    BaseMetaClass, SharedMethodsMixin, RSSFeedMixin, GetWatchersMixin, GetTagsMixin
):
    """The common class representing the page for either a Mod, Game, Engine or a Member. Mostly used to be inherited by
    those classes.

    Parameters
    -----------
    html : BeautifulSoup
        The page to be parsed.

    page_type : ThumbnailType
        The type of pages, this is passed down be the base class to help with the parsing of

    Attributes
    -----------
    name : str
        The name of the page
    profile : Profile
        The object with the content of the page's profile box.
    stats : Statistics
        The object containg stat data on the page such as views, followers, ect...
    style : Style
        The object containing data relevant to the type of the page, not valid for Engines. Multiplayer, singleplayer,
        ect...
    suggestions : List[Thumbnail]
        A list of thumbnail object representing moddb's suggestion of similar pages for the visiting member.
    files : List[Thumbnail]
        A list of thumbnails representing a possible partial list of all the files. To get a guaranteed full
        list either compare with statistics.files to see if the length of the list matches the number of files in the
        stats or use get_files, although that will still not return the whole list if there are multiple pages of
        files.
    articles : List[Thumbnail]
        A list of thumbnail objects representing articles present on the page. Usually 3 or 4 articles long.
    article :  PartialArticle
        A partial representation of the frong page article. This does not include things like comments or any
        of the sideba elements found in a full article. Can be parsed to return the complete Article object.
    tags : List[PartialTag]
        A list of partial tags. You can use `get_tags` and then use the name id to get the right one.
    imagebox : List[Thumbnail]
        A list of Thumbnail objects representing the image, videos and audio clips that are presented in the
        image box on the front page.
    embed : str
        The html necessary to embed the a widget of the page.
    rating : float
        A float out of ten representing the average rating for the page
    medias : List[Thumbnail]
        list of thumbnails representing all the combined media objects of a page (videos and images)
    summary : str
        Short description of the page, in plaintext.
    description : str
        The full description of the page, contains html
    plaintext : str
        Plaintext version of the full description
    """

    def __init__(self, html: BeautifulSoup, page_type: SearchCategory):
        super().__init__(html)
        self._type = page_type

        # boxes
        if page_type != SearchCategory.members:
            self.profile = Profile(html)

            try:
                self.stats = Statistics(html)
            except AttributeError:
                LOGGER.info(
                    "Entity '%s' has no stats (idk why, ask moddb)",
                    self.name,
                    exc_info=LOGGER.level >= logging.DEBUG,
                )
                self.stats = None

            if page_type != SearchCategory.engines:
                self.style = Style(html)

            # thumbnails
            self.suggestions = self._get_suggestions(html)

            # misc
            try:
                self.embed = html.find("input", type="text", class_="text textembed")["value"]
            except TypeError:
                self.embed = str(html.find_all("textarea")[1].a)

        try:
            self.files = self._get_files(html)
        except AttributeError:
            LOGGER.info(
                "'%s' '%s' has no files",
                self.__class__.__name__,
                self.name,
                exc_info=LOGGER.level >= logging.DEBUG,
            )
            self.files = []

        articles_raw = None
        try:
            raw = html.find("span", string="Articles") or html.find(
                "span", string="Related Articles"
            )
            articles_raw = raw.parent.parent.parent.find("div", class_="table")
            thumbnails = articles_raw.find_all("div", class_="row rowcontent clear")
            self.articles = [
                Thumbnail(
                    name=x.a["title"],
                    url=x.a["href"],
                    image=x.a.img["src"] if x.a.img else None,
                    summary=x.find("p").string,
                    date=x.find("time")["datetime"],
                    type=ThumbnailType.article,
                )
                for x in thumbnails
            ]
        except AttributeError:
            LOGGER.info(
                "'%s' '%s' has no article suggestions",
                self.__class__.__name__,
                self.name,
                exc_info=LOGGER.level >= logging.DEBUG,
            )
            self.articles = []

        # main page article
        if articles_raw:
            self.article = PartialArticle(articles_raw)
        else:
            self.article = None
            LOGGER.info(
                "'%s' '%s' has no front page article",
                self.__class__.__name__,
                self.name,
                exc_info=LOGGER.level >= logging.DEBUG,
            )

        try:
            raw_tags = html.find("form", attrs={"name": "tagsform"}).find_all("a")
            self.tags = [
                PartialTag(x.string, join(x["href"]), x["href"].split("/")[-1])
                for x in raw_tags
                if x.string is not None
            ]
        except AttributeError:
            self.tags = []
            LOGGER.info(
                "'%s' '%s' has no tags",
                self.__class__.__name__,
                self.name,
                exc_info=LOGGER.level >= logging.DEBUG,
            )

        # imagebox
        try:
            imagebox = html.find("ul", id="imagebox").find_all("li")[1:-2]
            self.imagebox = [
                Thumbnail(
                    name=x.a["title"],
                    url=x.a["href"],
                    image=x.a.img["src"],
                    type=ThumbnailType.media,
                )
                for x in imagebox
                if x.a
            ]
        except (AttributeError, TypeError):
            LOGGER.info(
                "'%s' '%s' has no images",
                self.__class__.__name__,
                self.name,
                exc_info=LOGGER.level >= logging.DEBUG,
            )
            self.imagebox = []

        try:
            self.rating = float(
                html.find("div", class_="score").find("meta", itemprop="ratingValue")["content"]
            )
        except AttributeError:
            self.rating = 0.0
            LOGGER.info(
                "'%s' '%s' is not rated",
                self.__class__.__name__,
                self.name,
                exc_info=LOGGER.level >= logging.DEBUG,
            )

        try:
            self._review_hash = html.find("form", class_="ratingform").find(
                "input", {"name": "hash"}
            )["value"]
        except AttributeError:
            self._review_hash = None

        self.medias = self._get_media(2, html=html)

        try:
            self.summary = html.find("meta", itemprop="description")["content"]
        except TypeError:
            self.summary = None
            LOGGER.info(
                "'%s' '%s' has no summary",
                self.__class__.__name__,
                self.name,
                exc_info=LOGGER.level >= logging.DEBUG,
            )

        self.description = str(html.find("div", id="profiledescription"))

        try:
            self.description = str(html.find("div", id="profiledescription"))
            self.plaintext = html.find("div", id="profiledescription").text
        except AttributeError:
            LOGGER.info(
                "'%s' '%s' has no extended description",
                self.__class__.__name__,
                self.name,
                exc_info=LOGGER.level >= logging.DEBUG,
            )
            self.description = None
            self.plaintext = None

    def _get_suggestions(self, html: BeautifulSoup) -> List[Thumbnail]:
        """Hidden method used to get the list of suggestions on the page. As with most things, this list of suggestions
        will be a list of Thumbnail objects that can be parsed individually.

        Parameters
        -----------
        html : BeautifulSoup
            The html page we are trying to parse the suggestions for

        Returns
        --------
        List[Thumbnail]
            The list of suggestions as thumbnails.
        """
        suggestions_raw = (
            html.find("span", string="You may also like")
            .parent.parent.parent.find("div", class_="table")
            .find_all("div", recursive=False)
        )
        suggestions = []
        for x in suggestions_raw:
            try:
                link = x.find("a", class_="image")
                suggestion_type = get_page_type(link["href"])
                suggestion = Thumbnail(
                    name=link["title"],
                    url=link["href"],
                    image=link.img["src"],
                    type=suggestion_type,
                )
                suggestions.append(suggestion)
            except (AttributeError, TypeError):
                pass

        return suggestions

    def _get_games(self, html: BeautifulSoup) -> List[Thumbnail]:
        """Used both for Teams and Engines, returns a list of games  present on the page
        as Thumbnail objects.

        Parameters
        ----------
        html : BeautifulSoup
            The html to extract the list of games from

        Returns
        -------
        List[Thumbnail]
            The list of games on the page as Thumbnail objects.
        """
        games_raw = html.find(string="Games").parent.parent.parent.parent.find_all(
            class_="row rowcontent clear"
        )
        games = []
        for x in games_raw:
            link = x.find("div", class_="content").h4.a
            image_url = link.parent.parent.parent.find("img")["src"]
            game = Thumbnail(
                name=link.string,
                url=link["href"],
                image=image_url,
                type=ThumbnailType.game,
            )
            games.append(game)

        return games

    def _get_files(self, html: BeautifulSoup) -> List[Thumbnail]:
        """Cache the files present on the page. Up to 5 files might be present

        Parameters
        -----------
        html : BeautifulSoup
            The page to cache the files from

        Returns
        -------
        List[Thumbnail]
            The list of file like thumbnails
        """

        files_raw = html.find(string="Files").parent.parent.parent.parent.find_all(
            class_="row rowcontent clear"
        )
        files = []
        for x in files_raw:
            link = x.find("div", class_="content").h4.a
            try:
                image_url = link.parent.parent.parent.find("img")["src"]
            except TypeError:
                image_url = None

            file = Thumbnail(
                name=link.string,
                url=link["href"],
                image=image_url,
                type=ThumbnailType.file,
            )
            files.append(file)

        return files

    def _get_engines(self, html: BeautifulSoup):
        """Hidden method to get the engines showed currently on the page as a list of engine like thumbnails. Takes
        an entire page of html and sorts it out

        Parameters
        -----------
        html : BeautifulSoup
            The page to cache the engines from

        Returns
        -------
        List[Thumbnail]
            The list of engine like thumbnail objects
        """
        engines_raw = html.find(string="Engines").parent.parent.parent.parent.find_all(
            class_="row rowcontent clear"
        )
        engines = []
        for x in engines_raw:
            link = x.find("div", class_="content").h4.a
            image_url = link.parent.parent.parent.find("img")["src"]
            engine = Thumbnail(
                name=link.string,
                url=link["href"],
                image=image_url,
                type=ThumbnailType.engine,
            )
            engines.append(engine)

        return engines

    def __repr__(self):
        return f"<{self.__class__.__name__} name={self.name}>"


class HardwareSoftwareMetaClass(
    BaseMetaClass, SharedMethodsMixin, RSSFeedMixin, GetWatchersMixin, GetTagsMixin
):
    """Shared class for Hardware and Software

    Attributes
    -----------
    description : str
        Description of the page
    profile : Profile
        The page's profile
    stats : Statistics
        The page's stats
    rating : float
        The rating of the item
    articles : List[Thumbnail]
        List of article type thumbnails from the recommended articles
    article : PartialArticle
        The partial article presented on the front page
    tags : List[PartialTag]
         A list of partial tags. You can use `get_tags` and then use the name id to get the right one.
    medias : List[Thumbnail]
        list of thumbnails representing all the combined media objects of a page (videos and images)
    suggestions : List[Thumbnail]
        list of suggested software/hardware type thumbnails
    """

    def __init__(self, html: BeautifulSoup):
        super().__init__(html)
        try:
            self.description = html.find("div", id="profiledescription").p.string
        except AttributeError:
            self.description = html.find("p", itemprop="description").string

        self.profile = Profile(html)

        try:
            self.stats = Statistics(html)
        except AttributeError:
            LOGGER.info(
                "Entity '%s' has no stats (idk why, ask moddb)",
                self.name,
                exc_info=LOGGER.level >= logging.DEBUG,
            )
            self.stats = None

        try:
            self.rating = float(
                html.find("div", class_="score").find("meta", itemprop="ratingValue")["content"]
            )
        except AttributeError:
            self.rating = 0.0
            LOGGER.info(
                "'%s' '%s' is not rated",
                self.profile.category.name,
                self.name,
                exc_info=LOGGER.level >= logging.DEBUG,
            )

        try:
            self._review_hash = html.find("form", class_="ratingform").find(
                "input", {"name": "hash"}
            )["value"]
        except AttributeError:
            self._review_hash = None

        articles_raw = None
        try:
            articles_raw = html.find("span", string="Related Articles").parent.parent.parent.find(
                "div", class_="table"
            )
            thumbnails = articles_raw.find_all("div", class_="row rowcontent clear")
            self.articles = [
                Thumbnail(
                    name=x.a["title"],
                    url=x.a["href"],
                    image=x.a.img["src"] if x.a.img else None,
                    summary=x.find("p").string,
                    date=x.find("time")["datetime"],
                    type=ThumbnailType.article,
                )
                for x in thumbnails
            ]
        except AttributeError:
            LOGGER.info(
                "'%s' '%s' has no article suggestions",
                self.profile.category.name,
                self.name,
                exc_info=LOGGER.level >= logging.DEBUG,
            )
            self.articles = []

        if articles_raw:
            self.article = PartialArticle(articles_raw)
        else:
            self.article = None
            LOGGER.info(
                "'%s' '%s' has no front page article",
                self.profile.category.name,
                self.name,
            )

        try:
            raw_tags = html.find("form", attrs={"name": "tagsform"}).find_all("a")
            self.tags = [
                PartialTag(x.string, join(x["href"]), x["href"].split("/")[-1])
                for x in raw_tags
                if x.string is not None
            ]
        except AttributeError:
            self.tags = []
            LOGGER.info(
                "Hardware '%s' has no tags", self.name, exc_info=LOGGER.level >= logging.DEBUG
            )

        self.medias = self._get_media(1, html=html)

        try:
            t = ThumbnailType[self.__class__.__name__.lower()]
            suggestions = html.find(
                "span", string="You may also like"
            ).parent.parent.parent.find_all("a", class_="image")
            self.suggestions = [
                Thumbnail(url=x["href"], name=x["title"], type=t, image=x.img["src"])
                for x in suggestions
            ]
        except AttributeError:
            LOGGER.info(
                "'%s' '%s' has no suggestions",
                self.__class__.__name__,
                self.name,
                exc_info=LOGGER.level >= logging.DEBUG,
            )
